/*
 * University of Illinois Open Source License
 * Copyright 2008-2011 Luthey-Schulten Group,
 * All rights reserved.
 * 
 * Developed by: Luthey-Schulten Group
 * 			     University of Illinois at Urbana-Champaign
 * 			     http://www.scs.uiuc.edu/~schulten
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the Software), to deal with 
 * the Software without restriction, including without limitation the rights to 
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
 * of the Software, and to permit persons to whom the Software is furnished to 
 * do so, subject to the following conditions:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimers.
 * 
 * - Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimers in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * - Neither the names of the Luthey-Schulten Group, University of Illinois at
 * Urbana-Champaign, nor the names of its contributors may be used to endorse or
 * promote products derived from this Software without specific prior written
 * permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL 
 * THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR 
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
 * OTHER DEALINGS WITH THE SOFTWARE.
 *
 * Author(s): Elijah Roberts
 */

#include <pthread.h>
#include "lm/Print.h"
#include "lm/thread/Thread.h"

namespace lm {
namespace thread {

Thread::Thread() throw(PthreadException)
:threadId(0),running(false)
{
    // Create the control mutex.
    PTHREAD_EXCEPTION_CHECK(pthread_mutex_init(&controlMutex, NULL));
}

Thread::~Thread() throw(PthreadException)
{
    PTHREAD_EXCEPTION_CHECK(pthread_mutex_destroy(&controlMutex));
}

void Thread::start() throw(PthreadException)
{
    //// BEGIN CRITICAL SECTION: controlMutex
    PTHREAD_EXCEPTION_CHECK(pthread_mutex_lock(&controlMutex));
    if (!running)
    {
        running=true;
        pthread_attr_t attr;
        PTHREAD_EXCEPTION_CHECK(pthread_attr_init(&attr));
        PTHREAD_EXCEPTION_CHECK(pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE));
        PTHREAD_EXCEPTION_CHECK(pthread_create(&threadId, &attr, &Thread::start_thread, this));
        PTHREAD_EXCEPTION_CHECK(pthread_attr_destroy(&attr));
        Print::printf(Print::DEBUG, "Started thread %u.", threadId);
    }
    PTHREAD_EXCEPTION_CHECK(pthread_mutex_unlock(&controlMutex));
    //// END CRITICAL SECTION: controlMutex
}

void * Thread::start_thread(void * obj)
{
    int ret = (reinterpret_cast<Thread *>(obj))->run();
    pthread_exit((void *)ret);
}

void Thread::stop() throw(PthreadException)
{
    bool waitForThread=false;

    //// BEGIN CRITICAL SECTION: controlMutex
    PTHREAD_EXCEPTION_CHECK(pthread_mutex_lock(&controlMutex));
    if (running)
    {
        Print::printf(Print::DEBUG, "Stopping thread %u.", threadId);
        running = false;
        waitForThread = true;
    }
    PTHREAD_EXCEPTION_CHECK(pthread_mutex_unlock(&controlMutex));
    //// END CRITICAL SECTION: controlMutex

    // Wait for the thread to stop, if we really stopped it.
    if (waitForThread)
    {
        // Try to wake up the thread in case it is sleeping.
        wake();

        // Join with the thread.
        void * ret;
        PTHREAD_EXCEPTION_CHECK(pthread_join(threadId, &ret));
        Print::printf(Print::DEBUG, "Thread %u stopped.", threadId);
    }
}

}
}
